/*
 * Copyright (c) 2010-2012 LAAS/CNRS
 * Arnaud Degroote - Wed Aug 11 2010
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netdb.h>

#include <tclserv_client/socket.h>

struct sock 
{
	int socket;
};

#ifndef NDEBUG
#define DPRINTF(format, ...) fprintf(stderr, format, __VA_ARGS__)
#else
#define DPRINTF(format, ...)
#endif

#define PROTECT_SYSCALL_FROM_EINTR(res,syscall) \
	while ((res = (syscall)) == -1 &&  (errno == EINTR)) {}

tclserv_client_socket_t 
tclserv_client_socket_init(const char* host, const char* port)
{
	tclserv_client_socket_t res;
	struct addrinfo hints;
	struct addrinfo *result, *rp;
	int s;
	int err;

	res = malloc(sizeof(*res));
	if (res == NULL) 
		return res;

	/* Obtain address(es) matching host/port */

	memset(&hints, 0, sizeof(struct addrinfo));
	hints.ai_family = AF_UNSPEC;    /* Allow IPv4 or IPv6 */
	hints.ai_socktype = SOCK_STREAM ; 
	hints.ai_flags = 0;
	hints.ai_protocol = 0;          

	s = getaddrinfo(host, port, &hints, &result);
	if (s != 0) {
		fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(s));
		goto clean;
	}

	for (rp = result; rp != NULL; rp = rp->ai_next) {
		res->socket = socket(rp->ai_family,
							 rp->ai_socktype,
							 rp->ai_protocol);
		if (res->socket == -1)
			continue;

		PROTECT_SYSCALL_FROM_EINTR(err,
			connect(res->socket, rp->ai_addr, rp->ai_addrlen));
		if (err != -1)
			break; /* success */

		PROTECT_SYSCALL_FROM_EINTR(err, close(res->socket));
	}

	freeaddrinfo(result);

	if (rp == NULL) {
		fprintf(stderr, "Could not connect\n");
		goto clean;
	}

	return res;

clean:
	free(res);
	return NULL;
}

ssize_t
tclserv_client_socket_write_request(tclserv_client_socket_t sock, const char* buf)
{
	int err;
	if (buf == NULL) return 0;
	if (sock == NULL) return -1;
	DPRINTF("DEBUG >>> writing %s\n", buf);
	PROTECT_SYSCALL_FROM_EINTR(err, 
		send(sock->socket, buf, strlen(buf), 0));
	if (err == -1) {
		int save_errno = errno;
		close(sock->socket);
		sock->socket=-1;
		errno = save_errno;
	}
	return err;
}

ssize_t
tclserv_client_socket_read_request(tclserv_client_socket_t sock, buf_t buf)
{
	return tclserv_client_socket_timed_read_request(sock, buf, NULL);
}



ssize_t 
tclserv_client_socket_timed_read_request(tclserv_client_socket_t sock, buf_t buf, struct timeval* timeout)
{
	fd_set rfds;
	ssize_t err;
	int nb_to_read = 0;
	char* p;
	size_t size_request;

	if (sock == NULL || buf == NULL) return -1;

	// wait for something to read
	FD_ZERO(&rfds);
	FD_SET(sock->socket, &rfds);


	PROTECT_SYSCALL_FROM_EINTR(err, 
				   select(sock->socket + 1, &rfds, NULL, NULL, timeout));
	if (err == -1) {
	    fprintf(stderr, "error in select : %s\n", strerror(errno));
		return -1;
	}
	
	if (err == 0) {
		errno = ETIME;
		return -1;
	}

	// get the len of the buffer
	if (ioctl(sock->socket, FIONREAD, &nb_to_read) == -1)
	{
		fprintf(stderr, "failed to call ioctl FIONREAD : %s\n", strerror(errno));
		return -1;
	}

	err = buf_reserve(buf, nb_to_read+1);
	if (err == -1) {
		fprintf(stderr, "Can't reserve enough memory for reading : %s\n", 
				strerror(errno));
		return -1;
	}

	PROTECT_SYSCALL_FROM_EINTR(err,
		 recv(sock->socket, buf->data, nb_to_read, MSG_PEEK));
	if (err == -1) {
		fprintf(stderr, "Failed to recv : %s\n", strerror(errno));
		return -1;
	}

	if (err == 0) {
		close(sock->socket);
		sock->socket = -1;
		fprintf(stderr, "Peer deconnection\n");
		return -1;
	}

	// look if we can find a rqst
	p = strstr(buf->data, "\r\n");
	if (p == NULL) { assert(0); } // XXX WTF here
	size_request = p - buf->data + 2;

	PROTECT_SYSCALL_FROM_EINTR(err,
		recv(sock->socket, buf->data, size_request, 0));
	buf->data[err] = 0;
	DPRINTF("DEBUG >>> reading %s\n", buf->data);
	buf->len = err;
	return err;
}

void 
tclserv_client_socket_destroy(tclserv_client_socket_t* sock)
{
	int err;
	if (*sock == NULL) 
		return;

	PROTECT_SYSCALL_FROM_EINTR(err, close((*sock)->socket));
	free(*sock);
	*sock = NULL;
}

int
tclserv_client_socket_is_connected(const tclserv_client_socket_t sock)
{
	return (sock->socket != -1);
}

int tclserv_client_socket_internal_descriptor(const tclserv_client_socket_t sock)
{
	return (sock->socket);
}
